/*
** Module   : TriggerData.java
** Abstract : This is a container that the database trigger manager uses for handling trigger
**             classes.
**
** Copyright (c) 2009-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------- Description --------------------------------
** 001 OM  20131206 Initial commit.
*/

package com.goldencode.p2j.persist.trigger;

import com.goldencode.p2j.util.DatabaseTrigger;

/**
 * This is a container for all information needed for a trigger callback. This class is
 * not mutable, the values passed to constructor can be later obtained using the individual
 * getters.
 * This class and all his methods are package-protected so it is not visible outside the trigger
 * package, in fact it can only be accessed by <code>TriggerBlockLayer</code> and
 * <code>DatabaseTriggerManager</code>.
 */
class TriggerData
{
   /** Trigger class that will be instantiated and executed. */
   private Class<? extends DatabaseTrigger> triggerClass;

   /** The container in which the trigger is declared. */
   private Object triggerContainer;

   /** Flag for disabling the schema trigger. */
   private boolean override;

   /** Flag that the trigger needs the old value of the buffer (WRITE only) */
   private boolean oldBuffer;

   /**
    * Builds an immutable object with data needed for a trigger to instantiate/execute.
    *
    * @param   triggerClass
    *          The class of the trigger.
    * @param   triggerContainer
    *          The container of the trigger.
    * @param   override
    *          Allows to block the schema trigger or execute before it.
    * @param   oldBuffer
    *          Some triggers need a buffer with old value.
    */
   TriggerData(Class<? extends DatabaseTrigger> triggerClass,
               Object triggerContainer,
               boolean override,
               boolean oldBuffer)
   {
      this.triggerClass = triggerClass;
      this.triggerContainer = triggerContainer;
      this.override = override;
      this.oldBuffer = oldBuffer;
   }

   /**
    * Obtain the trigger class.
    *
    * @return  The trigger class.
    */
   Class<? extends DatabaseTrigger> getTriggerClass()
   {
      return triggerClass;
   }

   /**
    * Obtain the trigger container.
    *
    * @return  The container object of the trigger.
    */
   Object getTriggerContainer()
   {
      return triggerContainer;
   }

   /**
    * Check if this trigger replaces the schema trigger (if any) or if it executes before.
    *
    * @return  The override flag.
    */
   boolean isOverride()
   {
      return override;
   }

   /**
    * Check if this trigger additionally need the old value of the buffer.
    *
    * @return  true if this trigger additionally need the old value of the buffer.
    */
   boolean hasOldBuffer()
   {
      return oldBuffer;
   }
}
