/*
** Module   : DatabaseTrigger.java
** Abstract : Abstract class to be extended by session database triggers.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------- Description --------------------------------
** 001 SVL 20130213 Created initial version.
** 002 OM  20131107 Added declaration for assign trigger methods and implementation of trigger() 
**                  method. Added replication support. Stubs for ON ASSIGN NEW VALUE form.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.persist.*;

/**
 * Abstract class to be extended by a session database trigger implementation. The trigger class
 * should implement a single function (<code>create</code>, <code>delete</code>, <code>find</code>
 * or some form of  <code>write</code>) which corresponds the database event type handled by the
 * trigger. Default implementations are no-ops.
 *
 * @param <Buf>
 *        The class of the buffer to which the database event served by this trigger is applied.
 */
public abstract class DatabaseTrigger<Buf extends Buffer>
{
   /**
    * Should be implemented by an ASSIGN trigger.
    *
    * @param   buf
    *          Buffer which contains the record that triggered the event.
    * @param   oldValueCopy
    *          A copy of the old value that was changed by assign statement. Even if this value
    *          can be overwritten, the change will be lost when exiting the trigger body block.
    */
   public void assign(Buf buf, BaseDataType oldValueCopy)
   {
   }

   /**
    * Should be implemented by an ASSIGN trigger. This overloaded method does not pass in the
    * old value.
    *
    * @param   buf
    *          Buffer which contains the record that triggered the event.
    */
   public void assign(Buf buf)
   {
   }

   /**
    * Should be implemented by a SCHEMA ASSIGN trigger with NEW form (no field reference).
    *
    * @param   newVal
    *          The value that triggered the event.
    * @param   oldValueCopy
    *          A copy of the old value that was changed by assign statement. Even if this value
    *          can be overwritten, the change will be lost when exiting the trigger body block.
    */
   public void assign(Undoable newVal, BaseDataType oldValueCopy)
   {
   }

   /**
    * Should be implemented by a SCHEMA ASSIGN trigger with NEW form (no field reference).
    * This overloaded method does not pass in the old value.
    *
    * @param   newVal
    *          The value that triggered the event.
    */
   public void assign(Undoable newVal)
   {
   }

   /**
    * Should be implemented by a CREATE trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void create(Buf buf)
   {
   }

   /**
    * Should be implemented by a FIND trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void find(Buf buf)
   {
   }

   /**
    * Should be implemented by a DELETE trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void delete(Buf buf)
   {
   }

   /**
    * Should be implemented by a WRITE trigger without OLD buffer specified.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void write(Buf buf)
   {
   }

   /**
    * Should be implemented by a WRITE trigger with OLD buffer specified.
    *
    * @param   newBuf
    *          Buffer which contains the new value of the record which triggered the event.
    * @param   oldBuf
    *          Buffer which contains the old value of the record which triggered the event.
    */
   public void write(Buf newBuf, Buf oldBuf)
   {
   }

   /**
    * Should be implemented by a REPLICATION-CREATE trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void replicationCreate(Buf buf)
   {
   }

   /**
    * Should be implemented by a REPLICATION-DELETE trigger.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void replicationDelete(Buf buf)
   {
   }

   /**
    * Should be implemented by a REPLICATION-WRITE trigger without OLD buffer specified.
    *
    * @param   buf
    *          Buffer which contains the record which triggered the event.
    */
   public void replicationWrite(Buf buf)
   {
   }

   /**
    * Should be implemented by a REPLICATION-WRITE trigger with OLD buffer specified.
    *
    * @param   newBuf
    *          Buffer which contains the new value of the record which triggered the event.
    * @param   oldBuf
    *          Buffer which contains the old value of the record which triggered the event.
    */
   public void replicationWrite(Buf newBuf, Buf oldBuf)
   {
   }

   /**
    * Execute the given trigger block.
    *
    * @param   blk
    *          The code for the trigger block body.
    */
   protected void trigger(Block blk)
   {
      // TODO: check if TransactionType.FULL is not the appropriate type
      BlockManager.topLevelBlock(BlockManager.TransactionType.SUB, BlockType.TRIGGER, blk);
   }
}
